"""
RasgoQL main class
"""
from typing import List, Type

import pandas as pd
import rasgotransforms as rtx

from .data import DataWarehouse
from .primitives import Dataset, SQLChain, TransformTemplate
from .version import __version__

class RasgoQL:
    """
    Entry class for rasgoql
    """
    __version__ = __version__

    def __init__(
            self,
            connection: Type[DataWarehouse],
            credentials: dict
        ):
        self.credentials = credentials
        self._dw = connection()
        self._dw.connect(credentials)

    def disconnect_dw(self):
        """
        Closes a connection to the Data Warehouse
        """
        self._dw.close_connection()

    def list_tables(self):
        """
        Returns a list of tables in your Data Warehouse
        """
        return self._dw.list_tables()

    def list_transforms(self) -> List[TransformTemplate]:
        """
        Returns a list of RasgoQL transforms
        """
        return rtx.serve_rasgo_transform_templates()

    def dataset(
            self,
            fqtn: str
        ) -> Dataset:
        """
        Returns a Dataset connected to the Cloud Data Warehouse
        """
        return Dataset(fqtn, self._dw)

    def define_transform(self, name: str) -> str:
        """
        Returns full details of a RasgoQL transform
        """
        udt: TransformTemplate = None
        for t in rtx.serve_rasgo_transform_templates():
            if t.name == name:
                udt = t
        if udt:
            return udt.define()
        raise ValueError(f'{name} is not a valid Tranform name. ' \
                           'Run `.list_transforms()` to see available transforms.')

    def query(
            self,
            sql: str
        ):
        """
        Execute a query against your Data Warehouse
        """
        return self._dw.execute_query(sql)

    def query_into_df(
            self,
            sql: str
        ) -> pd.DataFrame:
        """
        Execute a query against your Data Warehouse and return results in a pandas DataFrame
        """
        return self._dw.execute_query(sql, response='df')

    def sqlchain(
            self,
            fqtn: str
        ) -> SQLChain:
        """
        Returns a SQLChain connected to the Cloud Data Warehouse
        """
        return SQLChain(
            entry_table=Dataset(
                fqtn,
                self._dw
            ),
            dw=self._dw
        )



"""
Helpful sql utilities
"""
import random
import string


def cleanse_sql_data_type(
        dtype: str
    ) -> str:
    """
    Converts pandas data types to SQL compliant data type
    """
    if dtype.lower() in ["object", "text", "variant"]:
        return "string"
    else:
        return dtype.lower()

def cleanse_sql_name(
        name: str
) -> str:
    """
    Converts a string to a SQL compliant value
    """
    return name.replace(" ", "_").replace("-", "_").replace('"', '').replace(".","_").upper()

def is_scary_sql(
        sql: str
) -> bool:
    """
    Checks a SQL string for presence of injection keywords
    """
    if any(word in sql.upper() for word in SQL_INJECTION_KEYWORDS):
        return True
    return False

def is_restricted_sql(
        sql: str
) -> bool:
    """
    Checks a SQL string for presence of dangerous keywords
    """
    if any(word in sql.upper() for word in SQL_RESTRICTED_KEYWORDS):
        return True
    return False

def make_fqtn(
        database: str,
        schema: str,
        table: str
) -> str:
    """
    Accepts component parts and returns a fully qualified table string
    """
    return f"{database}.{schema}.{table}"

def parse_fqtn(
        fqtn: str
) -> tuple:
    """
    Accepts a possible fully qualified table string and returns its component parts
    """
    database = None
    schema = None
    table = fqtn
    if fqtn.count('.') == 2:
        database = fqtn.split(".")[0]
        schema = fqtn.split(".")[1]
        table = fqtn.split(".")[-1]
    elif fqtn.count('.') == 1:
        schema = fqtn.split(".")[0]
        table = fqtn.split(".")[-1]
    return database, schema, table

def random_table_name() -> str:
    return 'RQL_'+''.join(random.choice(string.ascii_uppercase) for x in range(10))

def validate_fqtn(
        fqtn: str
    ) -> bool:
    """
    Accepts a possible fully qualified table string and decides whether it is well formed
    """
    is_fqtn = False
    if fqtn.count('.') == 2:
        is_fqtn = True
    return is_fqtn

def wrap_table(
        parent_table: str
    )-> str:
    """
    Calculates a unique table string
    """
    prefix = 'RQL'
    suffix = ''.join(random.choice(string.ascii_uppercase) for x in range(5))
    if parent_table.startswith('RQL'):
        return f"{parent_table}_{suffix}"
    return f"{prefix}_{parent_table}_{suffix}"


SQL_RESTRICTED_CHARACTERS = [
    ' ', '-', ';'
]

SQL_INJECTION_KEYWORDS = [
    'DELETE',
    'TRUNCATE',
    'DROP',
    'ALTER',
    'UPDATE',
    'INSERT',
    'MERGE'
]

SQL_RESTRICTED_KEYWORDS = [
    'ACCOUNT','ALL','ALTER','AND','ANY','AS','BETWEEN','BY',
    'CASE','CAST','CHECK','COLUMN','CONNECT','CONNECTION','CONSTRAINT',
    'CREATE','CROSS','CURRENT','CURRENT_DATE','CURRENT_TIME','CURRENT_TIMESTAMP',
    'CURRENT_USER','DATABASE','DELETE','DISTINCT','DROP','ELSE','EXISTS','FALSE',
    'FOLLOWING','FOR','FROM','FULL','GRANT','GROUP','GSCLUSTER','HAVING','ILIKE',
    'IN','INCREMENT','INNER','INSERT','INTERSECT','INTO','IS','ISSUE','JOIN','LATERAL',
    'LEFT','LIKE','LOCALTIME','LOCALTIMESTAMP','MINUS','NATURAL','NOT','NULL','OF','ON',
    'OR','ORDER','ORGANIZATION','QUALIFY','REGEXP','REVOKE','RIGHT','RLIKE','ROW','ROWS',
    'SAMPLE','SCHEMA','SELECT','SET','SOME','START','TABLE','TABLESAMPLE','THEN','TO','TRIGGER',
    'TRUE','TRY_CAST','UNION','UNIQUE','UPDATE','USING','VALUES','VIEW','WHEN','WHENEVER','WHERE','WITH'
]



"""
Snowflake DataWarehouse classes
"""
import logging
import os
import re
from typing import List, Optional, Union

import json
import pandas as pd
from snowflake import connector

from rasgoql.errors import (
    DWConnectionError, DWQueryError,
    ParameterException, SQLException,
    TableAccessError, TableConflictException
)
from rasgoql.primitives.enums import check_response_type, check_table_type
from rasgoql.utils.creds import load_env, save_env
from rasgoql.utils.df import cleanse_sql_dataframe
from rasgoql.utils.sql import is_scary_sql, parse_fqtn

from .base import DataWarehouse, DWCredentials

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class SnowflakeCredentials(DWCredentials):
    """
    Snowflake Credentials
    """
    dw_type = 'snowflake'

    def __init__(
            self,
            account: str,
            user: str,
            password: str,
            role: str,
            warehouse: str,
            database: str,
            schema: str
        ):
        self.account = account
        self.user = user
        self.password = password
        self.role = role
        self.warehouse = warehouse
        self.database = database
        self.schema = schema

    def __repr__(self) -> str:
        return json.dumps(
            {
                "account": self.account,
                "user": self.user,
                "role": self.role,
                "warehouse": self.warehouse,
            }
        )

    @classmethod
    def from_env(
            cls,
            file_path: str = None
        ) -> 'SnowflakeCredentials':
        """
        Creates an instance of this Class from a .env file on your machine
        """
        load_env(file_path)
        return cls(
            os.getenv('snowflake_account'),
            os.getenv('snowflake_user'),
            os.getenv('snowflake_password'),
            os.getenv('snowflake_role'),
            os.getenv('snowflake_warehouse'),
            os.getenv('snowflake_database'),
            os.getenv('snowflake_schema')
        )

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        return {
            "account": self.account,
            "user": self.user,
            "password": self.password,
            "role": self.role,
            "warehouse": self.warehouse,
            "database": self.database,
            "schema": self.schema
        }

    def to_env(
            self,
            file_path: str
        ):
        """
        Saves credentials to a .env file on your machine
        """
        creds = f'snowflake_account={self.account}\n',
        creds += f'snowflake_user={self.user}\n',
        creds += f'snowflake_password={self.password}\n',
        creds += f'snowflake_role={self.role}\n',
        creds += f'snowflake_warehouse={self.warehouse}\n',
        creds += f'snowflake_database={self.database}\n',
        creds += f'snowflake_schema={self.schema}\n'
        return save_env(creds)


class SnowflakeDataWarehouse(DataWarehouse):
    """
    Snowflake DataWarehouse
    """
    dw_type = 'snowflake'
    credentials_class = SnowflakeCredentials

    def __init__(self):
        super().__init__()
        self.credentials: dict = None
        self.connection: connector.SnowflakeConnection = None

    # ---------------------------
    # Core Data Warehouse methods
    # ---------------------------

    def connect(
            self,
            credentials: Union[dict, SnowflakeCredentials]
        ):
        """
        Connect to Snowflake

        Params:
        `credentials`: dict:
            dict (or DWCredentials class) holding the connection credentials
        """
        if isinstance(credentials, SnowflakeCredentials):
            credentials = credentials.to_dict()

        # This allows you to track what queries were run by Rasgo in your history tab
        credentials.update({
            "application": "rasgo",
            "session_parameters": {
                "QUERY_TAG": "rasgo_open_source_sdk"
            }
        })
        try:
            self.credentials = credentials
            self.connection = connector.connect(**credentials)
        except connector.errors.DatabaseError as e:
            raise DWConnectionError(e)
        except connector.errors.ForbiddenError as e:
            raise DWConnectionError(e)
        except Exception as e:
            raise e

    def close_connection(self):
        """
        Close connection to Snowflake
        """
        try:
            if self.connection:
                self.connection.close()
            self.connection = None
            logger.info("Connection to Snowflake closed")
        except connector.errors.DatabaseError as e:
            raise DWConnectionError(e)
        except Exception as e:
            raise e

    def create(
            self,
            sql: str,
            fqtn: str,
            table_type: str = 'VIEW',
            acknowledge_overwrite: bool = False
        ):
        """
        Create a view or table from given SQL

        Params:
        `sql`: str:
            query that returns data (i.e. can be wrapped in a CREATE TABLE statement)
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
            Name for the new table
        `table_type`: str:
            One of values: [view, table]
        `acknowledge_overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        table_type = check_table_type(table_type)
        if self._table_exists(fqtn) and not acknowledge_overwrite:
            msg = f'A table or view named {fqtn} already exists. ' \
                   'If you are sure you want to overwrite it, ' \
                   'pass in acknowledge_overwrite=True and run this function again'
            raise TableConflictException(msg)
        query = f'CREATE OR REPLACE {table_type} {fqtn} AS {sql}'
        return self.execute_query(query)

    def execute_query(
            self,
            sql: str,
            response: str = 'tuple',
            acknowledge_risk: bool = False
        ):
        """
        Run a query against Snowflake and return all results

        `sql`: str:
            query text to execute
        `response`: str:
            Possible values: [dict, df, None]
        `acknowledge_risk`: bool:
            pass True when you know your SQL statement contains
            a potentially dangerous or data-altering operation
            and still want to run it against your DataWarehouse
        """
        response = check_response_type(response)
        if is_scary_sql(sql) and not acknowledge_risk:
            msg = 'It looks like your SQL statement contains a ' \
                  'potentially dangerous or data-altering operation.' \
                  'If you are positive you want to run this, ' \
                  'pass in acknowledge_risk=True and run this function again.'
            raise SQLException(msg)
        logger.debug('>>>Executing SQL:')
        logger.debug(sql)
        if response == 'DICT':
            return self._execute_dict_cursor(sql)
        if response == 'DF':
            return self._execute_df_cursor(sql)
        return self._execute_string(sql, ignore_results=(response == 'NONE'))

    def get_ddl(
            self,
            fqtn: str
        ) -> str:
        """
        Returns the create statement for a table or view

        `fqtn`: str:
            Fully-qualified Table Name (database.schema.table)
        """
        sql = f"SELECT GET_DDL('TABLE', '{fqtn}') AS DDL"
        query_response = self.execute_query(sql, response='dict')
        return query_response[0]['DDL']

    def get_object_details(
            self,
            fqtn: str
        ) -> tuple:
        """
        Return details of a table or view in Snowflake

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)

        Response:
            object exists: bool
            is rasgo object: bool
            object type: [table|view|unknown]
        """
        database, schema, table = parse_fqtn(fqtn)
        sql = f"SHOW OBJECTS LIKE '{table}' IN {database}.{schema}"
        result = self.execute_query(sql, response='dict')
        obj_exists = len(result) > 0
        is_rasgo_obj = False
        obj_type = 'unknown'
        if obj_exists:
            is_rasgo_obj = (result[0].get('comment') == 'rasgoql')
            obj_type = result[0].get('kind')
        return obj_exists, is_rasgo_obj, obj_type

    def get_schema(
            self,
            fqtn: str
        ) -> dict:
        """
        Return the schema of a table or view

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        """
        sql = f"DESC TABLE {fqtn}"
        query_response = self.execute_query(sql, response='dict')
        return query_response

    def list_tables(
            self,
            database: str = None,
            schema: str = None
        ) -> pd.DataFrame:
        """
        List all tables and views available in default namespace

        Params:
        `database`: str:
            override database
        `schema`: str:
            override schema
        """
        select_clause = "SELECT TABLE_NAME, " \
                        "TABLE_CATALOG||'.'||TABLE_SCHEMA||'.'||TABLE_NAME AS FQTN, " \
                        "CASE TABLE_TYPE WHEN 'BASE TABLE' THEN 'TABLE' ELSE TABLE_TYPE END AS TABLE_TYPE, " \
                        "ROW_COUNT, CREATED, LAST_ALTERED "
        from_clause = " FROM INFORMATION_SCHEMA.TABLES "
        if database:
            from_clause = f" FROM {database}.INFORMATION_SCHEMA.TABLES "
        where_clause = f" WHERE TABLE_SCHEMA = '{schema}'" if schema else ""
        sql = select_clause + from_clause + where_clause
        return self.execute_query(sql, response='df', acknowledge_risk=True)

    def preview(
            self,
            sql: str,
            limit: int = 10
        ) -> pd.DataFrame:
        """
        Returns 10 records into a pandas DataFrame

        Params:
        `sql`: str:
            SQL statment to run
        `limit`: int:
            Records to return
        """
        return self.execute_query(f'{sql} LIMIT {limit}', response='df', acknowledge_risk=True)

    def save_df(
            self,
            df: pd.DataFrame,
            fqtn: str,
            acknowledge_overwrite: bool = False
        ):
        """
        Creates a table in Snowflake from a pandas Dataframe

        Params:
        `df`: pandas DataFrame:
            DataFrame to upload
        `fqtn`: str:
            Fully-qualied table name (database.schema.table)
            Name for the new table
        `acknowledge_overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        if self._table_exists(fqtn) and not acknowledge_overwrite:
            msg = f'A table named {fqtn} already exists. ' \
                   'If you are sure you want to overwrite it, ' \
                   'pass in acknowledge_overwrite=True and run this function again'
            raise TableConflictException(msg)
        try:
            return connector.pandas_tools.write_pandas(
                conn=self.connection,
                df=cleanse_sql_dataframe(df),
                table_name=fqtn,
                quote_identifiers=False
            )
        except Exception as e:
            raise e

    # ---------------------------
    # Core Data Warehouse helpers
    # ---------------------------
    def _table_exists(
            self,
            fqtn: str
        ) -> bool:
        """
        Check for existence of fqtn in the Data Warehouse and return a boolean

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        """
        do_i_exist, _, _ = self.get_object_details(fqtn)
        return do_i_exist

    def _validate_namespace(
            self,
            namespace: str
        ):
        """
        Checks a namespace string for compliance with Snowflake format

        Params:
        `namespace`: str:
            namespace (database.schema.table)
        """
        # Does this match a 'string.string' pattern?
        if re.match(r'\w+\.\w+', namespace):
            return
        raise ParameterException("Snowflake namespaces should be format: DATABASE.SCHEMA")

    # --------------------------
    # Snowflake specific helpers
    # --------------------------
    def _execute_string(
            self,
            query: str, *,
            ignore_results: bool = False
        ) -> List[tuple]:
        """
        Execute a query string against the Data Warehouse connection and fetch all results
        """
        query_returns = []
        cursor = None
        try:
            for cursor in self.connection.execute_string(query, return_cursors=(not ignore_results)):
                for query_return in cursor:
                    query_returns.append(query_return)
            return query_returns
        except connector.errors.ProgrammingError as e:
            raise DWQueryError(e)
        except Exception as e:
            raise e
        finally:
            if cursor:
                cursor.close()

    def _execute_dict_cursor(
            self,
            query: str
        ) -> List[dict]:
        """
        Run a query string and return results in a Snowflake DictCursor

        PRO:
        Results are callable by column name
        for row in data:
            row['COL_NAME']

        CON:
        Query string must be a single statement (only one ;) or Snowflake returns an error
        """
        cursor = None
        try:
            cursor = self.connection.cursor(connector.DictCursor)
            query_return = cursor.execute(query).fetchall()
            return query_return
        except connector.errors.ProgrammingError as e:
            raise DWQueryError(e)
        except Exception as e:
            raise e
        finally:
            if cursor:
                cursor.close()

    def _execute_df_cursor(
            self,
            query: str,
            params: Optional[dict] = None
        ) -> pd.DataFrame:
        """
        Run a query string and return results in a pandas DataFrame
        """
        cursor = None
        try:
            cursor = self.connection.cursor()
            query_return = cursor.execute(query, params).fetch_pandas_all()
            return query_return
        except connector.errors.ProgrammingError as e:
            raise DWQueryError(e)
        except Exception as e:
            raise e
        finally:
            if cursor:
                cursor.close()



"""
DataWarehouse credential functions
"""
import os
import logging
import yaml

import dotenv

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def load_env(
        file_path: str = None
    ):
    """
    Loads a .env file, allowing contents to be callable by os.getenv()
    """
    dotenv.load_dotenv(file_path)

def load_yml(
        file_path: str
    ):
    """
    Loads a yml file and returns the contents
    """
    with open(file_path) as yml_file:
        contents = yaml.load(yml_file, Loader=yaml.Loader)
    return contents

def save_env(
        creds: str
    ):
    """
    Writes creds to a .env file
    """
    file_path = os.getcwd() + '/.env'
    with open(file_path, "w") as _file:
        _file.write(creds)
    return file_path

def save_yml(
        creds: str,
        file_path: str = None,
        acknowledge_overwrite: bool = True
    ):
    """
    Write creds to a .yml file
    """
    if file_path is None:
        file_path = os.getcwd()
        file_path += 'credentials.yml'

    if file_path[-1] == "/":
        file_path = file_path[:-1]

    if file_path.split(".")[-1] not in ['yaml', 'yml']:
        file_path += ".yml"

    if os.path.exists(file_path):
        if acknowledge_overwrite:
            logger.warning(f"Overwriting existing file {file_path}")
        else:
            raise FileExistsError(f'{file_path} already exists. If you wish to overwrite it, ' \
                                   'pass acknowledge_overwrite=True and run this function again.')

    with open(file_path, "w") as _yaml:
        yaml.dump(data=creds, Dumper=yaml.SafeDumper, stream=_yaml)



"""
Transform rendering methods
"""
import functools
import inspect
import logging
import re
from itertools import combinations, permutations, product
from typing import Callable, Dict, List, Optional

import jinja2
import pandas as pd
import rasgotransforms as rtx

from rasgoql.errors import TransformRenderingError
from .enums import check_table_type

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

RUN_QUERY_LIMIT = 100
JINJA_ENV = jinja2.Environment(extensions=['jinja2.ext.do', 'jinja2.ext.loopcontrols'])


def assemble_cte_chain(
        transforms: List['Transform'],
        table_type: str = None
    ):
    """
    Returns a nested CTE statement to render input Transforms
    """
    if table_type:
        table_type = check_table_type(table_type)

    create_stmt, final_select = '', ''

    # Handle single transform chains
    if len(transforms) == 1:
        t = transforms[0]
        if table_type in ['TABLE', 'VIEW']:
            create_stmt = f'CREATE OR REPLACE {table_type} {t.fqtn} AS \n'
        else:
            create_stmt = ''
        final_select = generate_transform_sql(
            t.name,
            t.arguments,
            t.source_table,
            None,
            t._dw
        )
        return create_stmt + final_select

    # Handle multi-transform chains
    t_list = []
    running_sql = None
    i = 1
    for t in transforms:
        logger.debug(f'Rendering transform {t.name} ({i} of {len(transforms)})')
        i += 1

        # Set final select & create statements from terminal transform
        if t == transforms[-1]:
            final_select = generate_transform_sql(
                t.name,
                t.arguments,
                t.source_table,
                running_sql,
                t._dw
            )
            if table_type in ['TABLE', 'VIEW']:
                create_stmt = f'CREATE OR REPLACE {table_type} {t.fqtn} AS \n'
            else:
                create_stmt = ''
        else:
            t_sql = generate_transform_sql(
                t.name,
                t.arguments,
                t.source_table,
                running_sql,
                t._dw
            )
            # running_sql tracks the CTE chain up to this transform,
            # this is consumed by the next transform if it needs to
            # access the chain's data mid-state (e.g. confirming columns exits)
            if len(t_list) == 0:
                running_sql = t_sql
            else:
                running_sql = 'WITH ' + ', \n'.join(t_list) + t_sql
            t_cte_str = f'{t.output_alias} AS (\n{t_sql}\n) '
            t_list.append(t_cte_str)
    return create_stmt + 'WITH ' + ', \n'.join(t_list) + final_select

def assemble_view_chain(
        transforms: List['Transform']
    ):
    """
    Returns a nested view statement to render input Transforms
    """
    view_list = []
    cte_list = []
    running_sql = None
    for t in transforms:
        t_sql = generate_transform_sql(
            t.name,
            t.arguments,
            t.source_table,
            running_sql,
            t._dw
            )
        if len(cte_list) == 0:
            running_sql = t_sql
        else:
            running_sql = 'WITH ' + ', \n'.join(cte_list) + t_sql
        t_cte_str = f'{t.output_alias} AS (\n{t_sql}\n) '
        t_view_str = f'CREATE OR REPLACE VIEW {t.fqtn} AS {t_sql};'
        cte_list.append(t_cte_str)
        view_list.append(t_view_str)
    return '\n'.join(view_list)

def generate_transform_sql(
        name: str,
        arguments: dict,
        source_table: str = None,
        running_sql: str = None,
        dw: 'DataWarehouse' = None
    ) -> str:
    """
    Returns the SQL for a Transform with applied arguments
    """
    templates = rtx.serve_rasgo_transform_templates()
    udt: 'TransformTemplate' = [t for t in templates if t.name == name][0]
    if not udt:
        raise TransformRenderingError(f'Cannot find a transform named {name}')
    try:
        source_code = udt.source_code
        if udt.name == 'apply':
            source_code = arguments.pop('sql')
            if not source_code:
                raise TransformRenderingError(f'Custom transform {udt.name} must provide ' \
                                               'the "sql" argument with template to apply')
        return render_template(name, source_code, arguments, source_table, running_sql, dw)
    except Exception as e:
        raise TransformRenderingError(e)

def render_template(
        name: str,
        source_code: str,
        arguments: dict,
        source_table: str,
        running_sql: str,
        dw: 'DataWarehouse'
    ) -> str:
    """
    Returns a SQL statement generated by applying arguments to a Jinja template
    """
    template_fns = _source_code_functions(dw, source_table, running_sql)
    template = JINJA_ENV.from_string(source_code)
    rendered = template.render(**arguments, **template_fns)
    if not rendered:
        raise TransformRenderingError(f'Rendering of transform {name} produced an empty string. ' \
                                       'Please check template arguments and try again.')
    return rendered

def _cleanse_template_symbol(
        symbol: str
    ) -> str:
    """
    Extra verbose function for clarity

    remove double quotes
    replace spaces and dashes with underscores
    cast to upper case
    delete anything that is not letters, numbers, or underscores
    if first character is a number, add an underscore to the beginning
    """
    symbol = symbol.strip()
    symbol = symbol.replace(' ', '_').replace('-', '_')
    symbol = symbol.upper()
    symbol = re.sub('[^A-Z0-9_]+', '', symbol)
    symbol = '_'+symbol if symbol[0].isdecimal() or not symbol else symbol
    return symbol

def _raise_exception(
        message: str
    ) -> None:
    """
    Raise an exception to return to users of a template
    """
    raise TransformRenderingError(message)

def _run_query(
        query: str,
        source_table: str = None,
        running_sql: str = None,
        dw: 'DataWarehouse' = None
    ) -> pd.DataFrame:
    """
    Jinja Func to materialize a chain as a temporary view before running a query
    """
    try:
        if running_sql:
            create_sql = f"CREATE OR REPLACE VIEW {source_table} AS {running_sql} LIMIT {RUN_QUERY_LIMIT}"
            dw.execute_query(create_sql, response='none', acknowledge_risk=True)
        return dw.execute_query(query, response='df', acknowledge_risk=True)
    except Exception as e:
        raise TransformRenderingError(e)
    finally:
        if running_sql:
            drop_sql = f"DROP VIEW IF EXISTS {source_table}"
            dw.execute_query(drop_sql, response='none', acknowledge_risk=True)

def _source_code_functions(
        dw: 'DataWarehouse',
        source_table: str = None,
        running_sql: str = None
    ) -> Dict[str, Callable]:
    """
    Get custom functions to add to the template parser
    """
    return {
        "run_query": functools.partial(_run_query, dw=dw, source_table=source_table, running_sql=running_sql),
        "cleanse_name": _cleanse_template_symbol,
        "raise_exception": _raise_exception,
        "itertools": {
            "combinations": combinations,
            "permutations": permutations,
            "product": product
        }
    }

def _gen_udt_func_signature(
        udt_func: Callable,
        transform: 'TransformTemplate'
    ) -> inspect.Signature:
    """
    Creates and returns a UDT param signature.

    This is shown documentation for the parameters when hitting shift tab in a notebook
    """
    sig = inspect.signature(udt_func)

    udt_params = []
    for t_arg in transform.arguments:
        p = inspect.Parameter(name=t_arg.get('name'), kind=inspect.Parameter.KEYWORD_ONLY)
        udt_params.append(p)

    op_name_param = inspect.Parameter(
        name='operation_name',
        kind=inspect.Parameter.KEYWORD_ONLY,
        annotation=Optional[str],
        default=None
    )
    udt_params.append(op_name_param)
    return sig.replace(parameters=udt_params)


def _gen_udt_func_docstring(
        transform: 'TransformTemplate'
    ) -> str:
    """
    Generate and return a docstring for a transform func
    with transform description, args, and return specified.
    """
    docstring = f"\n{transform.description}"

    docstring = f"{docstring}\n  Args:"
    for t_arg in transform.arguments:
        docstring = f"{docstring}\n    {t_arg.get('name')}: {t_arg.get('description')}"
    docstring = f"{docstring}\n    operation_name: Name to set for the operation"

    docstring = f"{docstring}\n\n  Returns:\n    Returns an new dataset with the referenced " \
                f"{transform.name!r} added to this dataset's definition"
    return docstring
